﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using SuperMap.Desktop.UI.Conversion;
using SuperMap.Data.Conversion;
using SuperMap.Data;
using System.Data.Odbc;
using System.Data.OleDb;
using DevExpress.XtraEditors;
using System.IO;

namespace SuperMap.Desktop.Samples
{
    public partial class FormDataImportTab : FormDataImport
    {
        public FormDataImportTab(bool? defaultFileTypeMDB)
        {
            InitializeComponent();
            m_fieldNames = new Dictionary<Int32, String>();
            if (defaultFileTypeMDB == null)
            {
                DefaultFileType = (FileType)FILETYPE_CUSTOM;
            }
            else
            {
                if (Convert.ToBoolean(defaultFileTypeMDB))
                {
                    DefaultFileType = (FileType)FILETYPE_MDB;
                }
                else
                {
                    DefaultFileType = (FileType)FILETYPE_EXCEL;
                }
            }
        }

        const int FILETYPE_MDB = 10001;
        const int FILETYPE_EXCEL = 10002;
        const int FILETYPE_CUSTOM = 10003;
        DataFileTypeInfo m_dftiMDB;
        DataFileTypeInfo m_dftiExcel;

        private Dictionary<Int32, String> m_fieldNames;
        // Initialize the file format which you want
        protected override void InitializeSupportFileTypes()
        {
            try
            {
                base.InitializeSupportFileTypes();

                // MDB
                m_dftiMDB = new DataFileTypeInfo();

                m_dftiMDB.Category = FileCategory.Tabular;
                m_dftiMDB.FileExt = "*.MDB";

                m_dftiMDB.FileType = (SuperMap.Data.Conversion.FileType)FILETYPE_MDB;
                m_dftiMDB.FileCaption = "Microsoft Access Files";

                m_dftiMDB.NameType = FileNameType.Name;

                m_dftiMDB.ParamPanel = new ImportSettingPanelMDB();
                m_dftiMDB.ImportSettingType = typeof(ImportSettingCSV);

                m_dftiMDB.ImportFunction = DoImportMDB;
                this.SupportFileTypes.Add(m_dftiMDB);

                // EXCEL
                m_dftiExcel = new DataFileTypeInfo();

                m_dftiExcel.Category = FileCategory.Tabular;
                m_dftiExcel.FileExt = "*.XLSX;*.XLS";

                m_dftiExcel.FileType = (SuperMap.Data.Conversion.FileType)FILETYPE_EXCEL;
                m_dftiExcel.FileCaption = "Microsoft Excel Files";

                m_dftiExcel.NameType = FileNameType.Name;

                m_dftiExcel.ParamPanel = new ImportSettingPanelExcel();
                m_dftiExcel.ImportSettingType = typeof(ImportSettingCSV);

                m_dftiExcel.ImportFunction = DoImportExcel;
                this.SupportFileTypes.Add(m_dftiExcel);

                // Cutom
                DataFileTypeInfo custom = new DataFileTypeInfo();
                custom.Category = FileCategory.Tabular;
                custom.FileExt = "*.custom";
                custom.FileType = (SuperMap.Data.Conversion.FileType)FILETYPE_CUSTOM;
                custom.FileCaption = "Custom Files";
                custom.NameType = FileNameType.Name;
                custom.ParamPanel = new ImportSettingCustomPanel();
                custom.ImportSettingType = typeof(ImportSettingCustom);
                custom.ImportFunction = DoImportCustom;
                this.SupportFileTypes.Add(custom);
            }
            catch (Exception ex)
            {
                Application.ActiveApplication.Output.Output(ex.StackTrace, InfoType.Exception);
            }
        }

        private SuperMap.Data.Conversion.ImportResult DoImportCustom(DataFileInfo dataFileInfo)
        {
            try
            {
                String path = dataFileInfo.ImportSetting.SourceFilePath;
                DataTable datatable = new DataTable();
                using (StreamReader reader = new StreamReader(path))
                {
                    String firstRow = reader.ReadLine();
                    String[] fieldValues = firstRow.Split(new char[] { ',' });
                    DataRow row = datatable.NewRow();
                    for (Int32 index = 0; index < fieldValues.Length; index++)
                    {
                        datatable.Columns.Add();
                        row[index] = fieldValues[index];
                    }
                    datatable.Rows.Add(row);
                    while (!reader.EndOfStream)
                    {
                        String value = reader.ReadLine();
                        if (!String.IsNullOrEmpty(value))
                        {
                            fieldValues = value.Split(new char[] { ',' });
                            row = datatable.NewRow();
                            for (Int32 index = 0; index < fieldValues.Length; index++)
                            {
                                row[index] = fieldValues[index];
                            }
                            datatable.Rows.Add(row);
                        }
                    }
                }
                ImportTable(dataFileInfo.ImportSetting.TargetDatasource, dataFileInfo.ImportSetting.TargetDatasetName, datatable, dataFileInfo);
            }
            catch (Exception ex)
            {
                Application.ActiveApplication.Output.Output(ex.StackTrace, InfoType.Exception);
            }
            return null;
        }

        private void ImportTable(Datasource targetDatasource, String tableName, DataTable datatable, DataFileInfo dataFileInfo)
        {
            try
            {
                DataTable tableScheme = datatable;
                DatasetVector targetDataset = CreateDataset(targetDatasource, tableName);
                CreateField(targetDataset, tableScheme);
                ImportTableToDataset(targetDataset, tableScheme, dataFileInfo);
                m_fieldNames.Clear();
            }
            catch (Exception ex)
            {
                Application.ActiveApplication.Output.Output(ex.StackTrace, InfoType.Exception);
            }
        }

        private void ImportTableToDataset(DatasetVector dataset, DataTable table, DataFileInfo dataFileInfo)
        {
            try
            {

                Recordset recordset = dataset.GetRecordset(true, CursorType.Dynamic);
                Int32 xfieldIndex = (dataFileInfo.ImportSetting as ImportSettingCustom).XFieldIndex;
                Int32 yfieldIndex = (dataFileInfo.ImportSetting as ImportSettingCustom).YFieldIndex;
                Int32 recordCount = table.Rows.Count;
                if (xfieldIndex < 0 || yfieldIndex < 0 || recordCount <= 0)
                {
                    return;
                }
                Int32 current = 0;
                Dictionary<String, Object> fieldValues = new Dictionary<String, Object>();
                Recordset.BatchEditor editor = recordset.Batch;
                editor.MaxRecordCount = 5000;
                editor.Begin();
                fieldValues.Clear();
                Object[] values = new Object[table.Columns.Count];
                for (Int32 rowIndex = 1; rowIndex < table.Rows.Count; rowIndex++)
                {
                    fieldValues.Clear();
                    for (int i = 0; i < table.Columns.Count; i++)
                    {
                        Object value = table.Rows[rowIndex][i];
                        FieldInfo info = dataset.FieldInfos[m_fieldNames[i]];
                        if (info.Type == FieldType.Text || info.Type == FieldType.WText)
                        {
                            value = (value == null) ? String.Empty : value.ToString();
                        }
                        else
                        {
                            value = (value.ToString() == "") ? null : value;
                        }
                        fieldValues.Add(m_fieldNames[i], value);
                    }
                    try
                    {
                        double x = Convert.ToDouble(table.Rows[rowIndex][xfieldIndex]);
                        double y = Convert.ToDouble(table.Rows[rowIndex][yfieldIndex]);
                        GeoPoint geopoint = new GeoPoint(new Point2D(x, y));
                        recordset.AddNew(geopoint, fieldValues);
                        geopoint.Dispose();
                        geopoint = null;
                    }
                    catch
                    {
                        //if build geopoint failed, do nothing
                    }
                    current++;
                    ImportSteppedEventArgs arg = new ImportSteppedEventArgs("import attribute table as a point dataset", " ", 100, current * 100 / recordCount, dataFileInfo.ImportSetting, recordCount, false);
                    this.UpdateImportProgress(arg);
                }
                editor.Update();
                recordset.Close();
                recordset.Dispose();
            }
            catch (Exception ex)
            {
                Application.ActiveApplication.Output.Output(ex.StackTrace, InfoType.Exception);
            }
        }

        private void CreateField(DatasetVector dataset, DataTable tableScheme)
        {
            try
            {
                m_fieldNames.Clear();
                for (Int32 i = 0; i < tableScheme.Columns.Count; i++)
                {
                    String name = tableScheme.Rows[0][i].ToString();
                    String fieldName = CreateField(dataset, name, null, 255);
                    if (!String.IsNullOrEmpty(fieldName))
                    {
                        m_fieldNames.Add(i, fieldName);
                    }
                }
            }
            catch (Exception ex)
            {
                Application.ActiveApplication.Output.Output(ex.StackTrace, InfoType.Exception);
            }
        }


        #region

        private SuperMap.Data.Conversion.ImportResult DoImportExcel(DataFileInfo dataFileInfo)
        {
            try
            {
                Datasource ds = dataFileInfo.ImportSetting.TargetDatasource;

                //String concString = "Provider=Microsoft.ACE.OLEDB.12.0;Data Source='" + serverFilename + "';Extended Properties='Excel 12.0;HDR=YES'";

                //+ "; Extended Properties=Excel 8.0;HDR=YES;IMEX=1"
                String concString = "Provider=Microsoft.ACE.OLEDB.12.0; Data Source=" + dataFileInfo.ImportSetting.SourceFilePath + "; Extended Properties=Excel 12.0 Xml";

                OleDbConnection conc = new OleDbConnection(concString);
                conc.Open();

                DataTable dt = conc.GetOleDbSchemaTable(OleDbSchemaGuid.Tables, new object[] { null, null, null, "TABLE" });

                List<String> invalidTableNames = new List<String>(dt.Rows.Count);

                foreach (DataRow row in dt.Rows)
                {
                    DataColumn dc = dt.Columns["TABLE_NAME"];
                    String tableName = row[dc].ToString();

                    if (tableName.Contains("$"))
                    {
                        if (invalidTableNames.Contains(tableName))
                        {
                            continue;
                        }
                        else
                        {

                        }
                    }
                    else
                    {
                        invalidTableNames.Add(tableName + "$");
                    }

                    ImportTable(ds, tableName, conc, dataFileInfo);
                }
                conc.Close();
            }
            catch (Exception ex)
            {
                Application.ActiveApplication.Output.Output(ex.StackTrace, InfoType.Exception);
            }
            return null;
        }
        
        // MDB function
        private SuperMap.Data.Conversion.ImportResult DoImportMDB(DataFileInfo dataFileInfo)
        {
            try
            {
                Datasource ds = dataFileInfo.ImportSetting.TargetDatasource;

                String concString = "Provider=Microsoft.Jet.OLEDB.4.0; Data Source=" + dataFileInfo.ImportSetting.SourceFilePath;

                OleDbConnection conc = new OleDbConnection(concString);
                conc.Open();

                DataTable dt = conc.GetOleDbSchemaTable(OleDbSchemaGuid.Tables, new object[] { null, null, null, "TABLE" });

                foreach (DataRow row in dt.Rows)
                {
                    DataColumn dc = dt.Columns["TABLE_NAME"];
                    String tableName = row[dc].ToString();

                    ImportTable(ds, tableName, conc, dataFileInfo);
                }
                conc.Close();
            }
            catch (Exception ex)
            {
                Application.ActiveApplication.Output.Output(ex.StackTrace, InfoType.Exception);
            }
            return null;
        }

        private DatasetVector CreateDataset(Datasource targetDatasource, String datasetName)
        {
            DatasetVector result = null;
            try
            {
                DatasetVectorInfo dtInfo = new DatasetVectorInfo();
                dtInfo.Name = targetDatasource.Datasets.GetAvailableDatasetName(datasetName); ;
                dtInfo.Type = DatasetType.Tabular;

                result = targetDatasource.Datasets.Create(dtInfo);
            }
            catch (Exception ex)
            {
                Application.ActiveApplication.Output.Output(ex.StackTrace, InfoType.Exception);
            }
            return result;
        }

        private void ImportTable(Datasource targetDatasource, String tableName, OleDbConnection conc, DataFileInfo dataFileInfo)
        {
            try
            {
                OleDbCommand command = new OleDbCommand("SELECT * FROM [" + tableName + "]", conc);
                DataSet dataset = new DataSet();
                OleDbDataAdapter objAdapter = new OleDbDataAdapter();
                objAdapter.SelectCommand = command;

                DataTableReader reader = null;

                try
                {
                    objAdapter.Fill(dataset);
                    reader = dataset.CreateDataReader();
                }
                catch
                {

                }
                if (reader == null ||
                    !reader.HasRows)
                {
                    return;
                }
                OleDbCommand cmdCount = new OleDbCommand("SELECT COUNT(*) FROM [" + tableName + "]", conc);
                int recordCount = Convert.ToInt32(cmdCount.ExecuteScalar());//Return the first row in query data

                DataTable tableScheme = reader.GetSchemaTable();

                DatasetVector targetDataset = CreateDataset(targetDatasource, tableName);
                CreateFields(targetDataset, tableScheme);

                ImportTable(targetDataset, reader, recordCount, tableName, dataFileInfo);
                m_fieldNames.Clear();
                reader.Close();
            }
            catch (Exception ex)
            {
                Application.ActiveApplication.Output.Output(ex.StackTrace, InfoType.Exception);
            }
        }

        private void CreateFields(DatasetVector dataset, DataTable tableScheme)
        {
            try
            {
                m_fieldNames.Clear();
                for (Int32 i = 0; i < tableScheme.Rows.Count; i++)
                {
                    DataRow row = tableScheme.Rows[i];
                    DataColumn columnName = tableScheme.Columns["ColumnName"];
                    DataColumn columnType = tableScheme.Columns["DataType"];
                    DataColumn columnSize = tableScheme.Columns["ColumnSize"];
                    String name = row[columnName].ToString();
                    Object objType = row[columnType];
                    int size = Convert.ToInt32(row[columnSize]);
                    String fieldName = CreateField(dataset, name, objType, size);
                    if (!String.IsNullOrEmpty(fieldName))
                    {
                        m_fieldNames.Add(i, fieldName);
                    }
                }
            }
            catch (Exception ex)
            {
                Application.ActiveApplication.Output.Output(ex.StackTrace, InfoType.Exception);
            }
        }

        private FieldType DbType2FieldType(Object dbTypeObject, int size)
        {
            FieldType fieldType = FieldType.Text;

            Type dbType = dbTypeObject as Type;
            if (dbType == typeof(System.Int32))
            {
                fieldType = FieldType.Int32;
            }
            else if (dbType == typeof(System.Int64))
            {
                fieldType = FieldType.Int64;
            }
            else if (dbType == typeof(System.Boolean))
            {
                fieldType = FieldType.Boolean;
            }
            else if (dbType == typeof(System.Byte))
            {
                fieldType = FieldType.Byte;
            }
            else if (dbType == typeof(System.Int16))
            {
                fieldType = FieldType.Int16;
            }
            else if (dbType == typeof(System.Single))
            {
                fieldType = FieldType.Single;
            }
            else if (dbType == typeof(System.Double))
            {
                fieldType = FieldType.Double;
            }
            else if (dbType == typeof(System.String))
            {
                fieldType = FieldType.Text;
            }
            else if (dbType == typeof(System.Array))
            {
                fieldType = FieldType.LongBinary;
            }
            else if (dbType == typeof(System.Char))
            {
                fieldType = FieldType.Text;
            }
            else if (dbType == typeof(System.DateTime))
            {
                fieldType = FieldType.DateTime;
            }
            else
            {
                fieldType = FieldType.Text;
            }

            return fieldType;
        }

        private String CreateField(DatasetVector dataset, String fieldName, Object oleDbType, int size)
        {
            String result = String.Empty;
            try
            {
                FieldInfo fieldInfo = new FieldInfo();
                fieldInfo.Caption = fieldName;
                fieldInfo.Type = DbType2FieldType(oleDbType, size);
                if ((fieldInfo.Type == FieldType.WText || fieldInfo.Type == FieldType.Text) &&
                    size <= 0)
                {
                    fieldInfo.MaxLength = 255;
                }

                fieldInfo.Name = dataset.GetAvailableFieldName(fieldName);
                Int32 index = dataset.FieldInfos.Add(fieldInfo);
                if (index >= 0)
                {
                    result = dataset.FieldInfos[index].Name;
                }
            }
            catch (Exception ex)
            {
                Application.ActiveApplication.Output.Output(ex.StackTrace, InfoType.Exception);
            }
            return result;
        }

        private void ImportTable(DatasetVector dataset,
                                                    DataTableReader reader,
                                                    int recordCount,
                                                    String tableName,
                                                    DataFileInfo dataFileInfo)
        {
            try
            {
                Recordset recordset = dataset.GetRecordset(true, CursorType.Dynamic);
                int current = 0;
                Dictionary<String, Object> fieldValues = new Dictionary<String, Object>(reader.FieldCount);
                // Get the batch update objects that correspond to recordset

                Recordset.BatchEditor editor = recordset.Batch;

                // Set the records number
                editor.MaxRecordCount = 1000000;

                editor.Begin();

                int maxIntervalTime = 100;// Update in 0.1 second

                DateTime start = DateTime.Now;

                while (reader.Read())
                {
                    fieldValues.Clear();
                    Object[] values = new Object[reader.FieldCount];
                    reader.GetValues(values);
                    for (int i = 0; i < reader.FieldCount; i++)
                    {
                        Object value = null;
                        FieldInfo info = dataset.FieldInfos[m_fieldNames[i]];
                        if (info.Type == FieldType.Text || info.Type == FieldType.WText)
                        {
                            value = (values[i] == null) ? String.Empty : values[i].ToString();
                        }
                        else
                        {
                            value = (values[i].ToString() == "") ? null : values[i];
                        }
                        fieldValues.Add(m_fieldNames[i], value);
                    }
                    recordset.AddNew(null, fieldValues);
                    current++;

                    TimeSpan span = DateTime.Now - start;
                    if (span.TotalMilliseconds > maxIntervalTime ||
                        current >= recordCount)
                    {
                        start = DateTime.Now;

                        ImportSteppedEventArgs arg = new ImportSteppedEventArgs("Import Attribute", "Import:" + tableName, 100, current * 100 / recordCount, dataFileInfo.ImportSetting, recordCount, false);
                        this.UpdateImportProgress(arg);
                    }
                }

                editor.Update();

                recordset.Close();
                recordset.Dispose();
            }
            catch (Exception ex)
            {
                Application.ActiveApplication.Output.Output(ex.StackTrace, InfoType.Exception);
            }
        }

        #endregion
    }
}
